unit U_Graph;

interface

uses SysUtils, U_adt;

const MAXKNOTEN = 10;
      MAXKANTEN = 100;
      MAXENTFERNUNG = maxInt div 2;
type
  TKnoten= char;
  TKante = record
                Start, Ziel: char;
                Gewicht: integer;
           end;
  TWeg = record
           Anzahl: integer;
           Kanten: array[1..MAXKANTEN] of TKante;
         end;

  TAdjazenzmatrix = array[1..MAXKNOTEN,1..MAXKNOTEN] of integer;

  TAdjazenzliste = array[1..MAXKNOTEN] of TLinList;

  TListenElement = class(TElement)
    Knotennummer: integer;
    Distanz: integer;
    constructor create(nr, dst: integer);
  end;

  TKnotenElement = class(TElement)
    Inhalt: integer;
    constructor create(i: integer);
  end;

  TGraph = class

                Knotenanzahl: integer;
                Kantenanzahl: integer;
                Knoten: array[1..MAXKNOTEN] of TKnoten;
                Kanten: array[1..MAXKANTEN] of TKante;

                constructor create;

                procedure Oeffnen(Filename: string);
                procedure Speichern(Filename: string);
                function SucheEulerkreis: TWeg;
                function PruefeEulerkreis: boolean;
                function Adjazenzmatrix: TAdjazenzmatrix;
                function Adjazenzliste: TAdjazenzliste;
                function Tiefensuche: TWeg;
                function Tiefensuche_iterativ: TWeg;
                function Breitensuche_iterativ: TWeg;
                function Hamilton: TWeg;
                function TSP: TWeg;
                function Dijkstra(Startknoten: integer): TWeg;
           end;


implementation

constructor TListenElement.create(nr, dst: integer);
begin
  inherited create;
  Knotennummer:= nr;
  Distanz:= dst;
end;


constructor TKnotenelement.create(i: integer);
begin
  inherited create;
  inhalt:= i;
end;

constructor TGraph.Create;
var i: integer;
begin
  inherited create;
  for i:=1 to MAXKNOTEN
  do Knoten[i]:= ' ';
  for i:= 1 to MAXKANTEN
  do begin
       Kanten[i].Start:= ' ';
       Kanten[i].Ziel:= ' ';
       Kanten[i].Gewicht:= MAXENTFERNUNG;
     end;
end;


procedure TGraph.Oeffnen(Filename: string);
var Datei: Textfile;
    i: integer;
    str: string;
begin
  assignfile(Datei,Filename);
  Reset(Datei);
  Readln(Datei,Knotenanzahl);
  for i:= 1 to Knotenanzahl
  do begin
       Readln(Datei,str);
       Knoten[i]:= str[1];
     end;
  Readln(Datei);
  Readln(Datei,Kantenanzahl);
  for i:= 1 to Kantenanzahl
  do begin
       readln(Datei,str);
       Kanten[i].Start:= str[1];
       Kanten[i].Ziel:= str[2];
       Kanten[i].Gewicht:= StrToInt(copy(str,4,10));
     end;
  closefile(Datei);
end;

procedure TGraph.Speichern(Filename: string);
var Datei: Textfile;
    i: integer;
begin
  assignfile(Datei,Filename);
  ReWrite(Datei);
  writeln(Datei,Knotenanzahl);
  for i:= 1 to Knotenanzahl
  do writeln(Datei,Knoten[i]);
  writeln(Datei,'---------------------');
  writeln(Datei,Kantenanzahl);
  for i:= 1 to Kantenanzahl
  do writeln(Datei,Kanten[i].Start+Kanten[i].Ziel+','+IntToStr(Kanten[i].Gewicht));
  closefile(Datei);
end;

function TGraph.SucheEulerkreis: TWeg;
var AktKnoten: TKnoten;
    KanteBesucht: array[1..MAXKANTEN] of boolean;
    AlleKantenBesucht: boolean;
    KantenNr: integer;
    i: integer;
begin
  result.Anzahl:= 0;
  for i:= 1 to MAXKANTEN
  do KanteBesucht[i]:= false;

  AktKnoten := Knoten[1];
  AlleKantenBesucht := (Kantenanzahl=0);
  While AlleKantenBesucht=false
  do begin
       { Suche Kante, welche vom aktuellen Knoten noch abgeht und nicht besucht wurde }
       KantenNr:= 0;
       for i := 1 to Kantenanzahl
       do begin
            if (KanteBesucht[i]=false) and
               ((Kanten[i].Start=AktKnoten) or (Kanten[i].Ziel=AktKnoten))
            then KantenNr:= i;
          end;
       if KantenNr > 0
       then begin      { Der Weg lt sich vom AktKnoten aus weiterfhren }
              inc(result.Anzahl);
              result.Kanten[result.Anzahl].Start:= AktKnoten;
              if AktKnoten<>Kanten[KantenNr].Ziel
              then AktKnoten:= Kanten[KantenNr].Ziel
              else AktKnoten:= Kanten[KantenNr].Start;
              result.Kanten[result.Anzahl].Ziel:= AktKnoten;
              KanteBesucht[KantenNr]:= true;
            end
       else begin      { Der Aktuelle Weg ist ein unvollstndiger Rundweg }
              { Starte den Rundweg von einem anderen Knoten ausgehend }
              result.Kanten[result.Anzahl+1]:= result.Kanten[1];
              for i:= 1 to result.Anzahl
              do result.Kanten[i]:= result.Kanten[i+1];
              AktKnoten:= result.Kanten[result.Anzahl].Ziel;
            end;
       { Testen, ob alle Kanten besucht sind }
       AlleKantenBesucht:= true;
       for i:= 1 to Kantenanzahl
       do if KanteBesucht[i]=false
          then AlleKantenBesucht:= false;
     end;
end;

function TGraph.PruefeEulerkreis: boolean;
var Knotennr, Kantennr: integer;
    Anzahl: integer;
begin
  result:= true;
  for Knotennr:= 1 to Knotenanzahl
  do begin
       Anzahl:= 0;
       for Kantennr:= 1 to Kantenanzahl
       do begin
            if (Kanten[Kantennr].Start=Knoten[Knotennr]) or
               (Kanten[Kantennr].Ziel=Knoten[Knotennr])
            then inc(Anzahl)
          end;
       if Anzahl mod 2 <> 0 then result:= false;
     end;
end;

function TGraph.Adjazenzmatrix: TAdjazenzmatrix;
var x,y,k: integer;
begin
  for x:= 1 to MAXKNOTEN
  do begin
       for y:= 1 to MAXKNOTEN
       do begin
            if x<>y then result[x,y]:= MAXENTFERNUNG
                    else result[x,y]:= 0;
            for k:= 1 to Kantenanzahl
            do if ((Kanten[k].Start=Knoten[x]) and (Kanten[k].Ziel=Knoten[y])) or
                  ((Kanten[k].Ziel=Knoten[x]) and (Kanten[k].Start=Knoten[y]))
               then result[x,y]:= Kanten[k].Gewicht;
          end;
     end;
end;

function TGraph.Adjazenzliste: TAdjazenzliste;
var x,y,k: integer;
begin
  for x:= 1 to Knotenanzahl
  do begin
       result[x]:= TLinList.create;
       for y:= 1 to Knotenanzahl
       do begin
            for k:= 1 to Kantenanzahl
            do if ((Kanten[k].Start=Knoten[x]) and (Kanten[k].Ziel=Knoten[y])) or
                  ((Kanten[k].Ziel=Knoten[x]) and (Kanten[k].Start=Knoten[y]))
               then result[x].EinfuegenNach(TListenElement.create(y,Kanten[k].Gewicht));
          end;
     end;
end;


function TGraph.Tiefensuche: TWeg;
var besucht: array[1..MAXKNOTEN] of boolean;
    Matrix: TAdjazenzmatrix;
    i: integer;
    Weg: TWeg;

  procedure SucheInDieTiefe(AktKnoten: integer);
  var Zielknoten: integer;
  begin
    for Zielknoten:= 1 to Knotenanzahl
    do if (not besucht[Zielknoten]) and (Matrix[AktKnoten,Zielknoten]>0) and (Matrix[AktKnoten,Zielknoten]<MAXENTFERNUNG)
       then begin
              inc(Weg.Anzahl);
              Weg.Kanten[Weg.Anzahl].Start:= Knoten[AktKnoten];
              Weg.Kanten[Weg.Anzahl].Ziel:= Knoten[Zielknoten];
              besucht[Zielknoten]:= true;
              SucheInDieTiefe(Zielknoten);
            end;
  end;

begin
  for i:= 1 to MAXKNOTEN
  do besucht[i]:= false;
  Weg.Anzahl:= 0;
  Matrix:= Adjazenzmatrix;

  for i:= 1 to Knotenanzahl
  do if not besucht[i]
     then begin
            besucht[i]:= true;
            SucheInDieTiefe(i);
          end;
  result:= Weg;
end;

function TGraph.Tiefensuche_iterativ: TWeg;
var Startknoten, Aktknoten, Zielknoten, i: integer;
    Stack: TStack;
    besucht: array[1..MAXKNOTEN] of boolean;
    Matrix: TAdjazenzmatrix;
begin
  Stack:= TStack.Create;
  for i:= 1 to Knotenanzahl
  do besucht[i]:= false;
  result.Anzahl:= 0;
  Matrix:= Adjazenzmatrix;
  for Startknoten:= 1 to Knotenanzahl
  do if not besucht[Startknoten]
     then begin
             besucht[Startknoten]:= true;
             Stack.push(TKnotenelement.create(Startknoten));
             while not Stack.empty
             do begin
                   Aktknoten:= (Stack.top as TKnotenelement).inhalt;
                   Zielknoten:= -1;
                   for i:= Knotenanzahl downto 1
                   do begin
                        if (not besucht[i]) and (Matrix[Aktknoten,i]>0) and (Matrix[Aktknoten,i]<MAXENTFERNUNG)
                        then Zielknoten:= i;
                      end;
                   if Zielknoten > -1
                   then begin
                          Stack.push(TKnotenelement.create(Zielknoten));
                          inc(result.Anzahl);
                          result.Kanten[result.Anzahl].Start:= Knoten[AktKnoten];
                          result.Kanten[result.Anzahl].Ziel:= Knoten[Zielknoten];
                          besucht[Zielknoten]:= true;
                        end
                   else Stack.pop;
                end;
          end;
  Stack.free;
end;

function TGraph.Breitensuche_iterativ: TWeg;
var Startknoten, Aktknoten, Zielknoten, i: integer;
    Queue: TQueue;
    besucht: array[1..MAXKNOTEN] of boolean;
    Matrix: TAdjazenzmatrix;
begin
  Queue:= TQueue.Create;
  for i:= 1 to Knotenanzahl
  do besucht[i]:= false;
  result.Anzahl:= 0;
  Matrix:= Adjazenzmatrix;
  for Startknoten:= 1 to Knotenanzahl
  do if not besucht[Startknoten]
     then begin
             besucht[Startknoten]:= true;
             Queue.enqueue(TKnotenelement.create(Startknoten));
             while not Queue.empty
             do begin
                   Aktknoten:= (Queue.front as TKnotenelement).inhalt;
                   Zielknoten:= -1;
                   for i:= Knotenanzahl downto 1
                   do begin
                        if (not besucht[i]) and (Matrix[Aktknoten,i]>0) and (Matrix[Aktknoten,i]<MAXENTFERNUNG)
                        then Zielknoten:= i;
                      end;
                   if Zielknoten > -1
                   then begin
                          Queue.enqueue(TKnotenelement.create(Zielknoten));
                          inc(result.Anzahl);
                          result.Kanten[result.Anzahl].Start:= Knoten[AktKnoten];
                          result.Kanten[result.Anzahl].Ziel:= Knoten[Zielknoten];
                          besucht[Zielknoten]:= true;
                        end
                   else Queue.dequeue;
                end;
          end;
  Queue.free;
end;



function TGraph.Hamilton: TWeg;
var Matrix: TAdjazenzmatrix;
    Kanten: TWeg;
    Ergebnis: TWeg;
    BesuchteKnoten: array[1..MAXKNOTEN] of boolean;

    procedure Initialisierung;
    var i: integer;
    begin
      Matrix:= Adjazenzmatrix;
      for i:= 1 to Knotenanzahl
      do BesuchteKnoten[i]:= false;

      Kanten.Anzahl:= 0;
    end;

    procedure Suche_rekursiv(Nr: integer);
    var i: integer;
        fertig: boolean;
    begin
      BesuchteKnoten[Nr]:= true;
      fertig:= true;
      for i:= 1 to KnotenAnzahl
      do if not BesuchteKnoten[i]
         then fertig:= false;
      if fertig
      then begin
             if (Matrix[Nr,1]>0) and (Matrix[Nr,1]<MAXENTFERNUNG) 
             then begin
                    Ergebnis:= Kanten;
                    inc(Ergebnis.Anzahl);
                    Ergebnis.Kanten[Ergebnis.Anzahl].Start:= Knoten[Nr];
                    Ergebnis.Kanten[Ergebnis.Anzahl].Ziel:= Knoten[1];
                  end;
           end
      else begin
             for i:= 1 to KnotenAnzahl
             do if (Matrix[Nr,i]>0) and (Matrix[Nr,i]<MAXENTFERNUNG) and (not BesuchteKnoten[i])
                then begin
                       inc(Kanten.Anzahl);
                       Kanten.Kanten[Kanten.Anzahl].Start:= Knoten[Nr];
                       Kanten.Kanten[Kanten.Anzahl].Ziel:= Knoten[i];
                       Suche_rekursiv(i);
                       dec(Kanten.Anzahl);
                     end;
           end;
      BesuchteKnoten[Nr]:= false;
    end;
begin
  Initialisierung;
  Suche_rekursiv(1);
  result:= Ergebnis;
end;



function TGraph.TSP: TWeg;
var Matrix: TAdjazenzmatrix;
    AktWeg, MinWeg: TWeg;
    BesuchteKnoten: array[1..MAXKNOTEN] of boolean;
    MinLaenge, AktLaenge: integer;

    procedure Initialisierung;
    var i: integer;
    begin
      Matrix:= Adjazenzmatrix;
      for i:= 1 to Knotenanzahl
      do BesuchteKnoten[i]:= false;
      aktWeg.Anzahl:= 0;
      MinLaenge:= maxInt;
      AktLaenge:= 0;
    end;

    procedure Suche_rekursiv(Nr: integer);
    var i: integer;
        fertig: boolean;
    begin
      BesuchteKnoten[Nr]:= true;
      fertig:= true;
      for i:= 1 to KnotenAnzahl
      do if not BesuchteKnoten[i]
         then fertig:= false;
      if fertig
      then begin
             if (Matrix[Nr,1]>0) and (Matrix[Nr,1]<MAXENTFERNUNG)
             then begin
                    inc(aktWeg.Anzahl);
                    aktWEg.Kanten[aktWeg.Anzahl].Start:= Knoten[Nr];
                    aktWeg.Kanten[aktWeg.Anzahl].Ziel:= Knoten[1];
                    if aktLaenge+Matrix[Nr,1]<minLaenge
                    then begin
                           minLaenge:= aktLaenge+Matrix[Nr,1];
                           minWeg:= aktWeg;
                         end;
                    dec(aktWeg.Anzahl);

                  end;
           end
      else begin
             for i:= 1 to KnotenAnzahl
             do if (Matrix[Nr,i]>0) and (Matrix[Nr,i]<MAXENTFERNUNG) and (not BesuchteKnoten[i])
                then begin
                       inc(aktWeg.Anzahl);
                       aktWeg.Kanten[aktWeg.Anzahl].Start:= Knoten[Nr];
                       aktWeg.Kanten[aktWeg.Anzahl].Ziel:= Knoten[i];
                       inc(AktLaenge,Matrix[Nr,i]);
                       Suche_rekursiv(i);
                       dec(aktWeg.Anzahl);
                       dec(AktLaenge,Matrix[Nr,i]);
                     end;
           end;
      BesuchteKnoten[Nr]:= false;
    end;
begin
  Initialisierung;
  Suche_rekursiv(1);
  result:= minWeg;
end;


function TGraph.Dijkstra(Startknoten: integer): TWeg;
var Vorgaenger, Distanz: array[1..MAXKNOTEN] of integer;
    Erledigt: array[1..MAXKNOTEN] of boolean;
    Matrix: TAdjazenzmatrix;
    AktKnoten, MinKnoten: integer;
    i: integer;

    function Knoten_mit_kleinster_Distanz: integer;
    var i: integer;
        mindistanz: integer;
    begin
      mindistanz:= maxEntfernung;
      result:= -1;
      for i:= 1 to Knotenanzahl
      do if (Distanz[i] < mindistanz) and (not Erledigt[i])
         then begin
                result:= i;
                mindistanz:= Distanz[i];
              end;
    end;

begin
  Matrix:= Adjazenzmatrix;
  for i:= 1 to Knotenanzahl
  do begin
       Distanz[i]:= maxEntfernung;
       Vorgaenger[i]:= -1;
       Erledigt[i]:= false;
     end;
  Distanz[Startknoten]:= 0;
  for i:= 1 to Knotenanzahl
  do begin
       MinKnoten:= Knoten_mit_kleinster_Distanz;
       If MinKnoten <> -1
       then begin
              Erledigt[MinKnoten]:= true;
              for AktKnoten:= 1 to Knotenanzahl
              do if (not Erledigt[AktKnoten]) and (Distanz[MinKnoten]+Matrix[MinKnoten,AktKnoten] < Distanz[AktKnoten])
                 then begin
                        Distanz[AktKnoten]:= Distanz[MinKnoten]+Matrix[MinKnoten,AktKnoten];
                        Vorgaenger[AktKnoten]:= MinKnoten;
                      end;
            end;
     end;
  result.Anzahl:= 0;
  for i:= 1 to Knotenanzahl
  do begin
       if i <> Startknoten
       then begin
              inc(result.Anzahl);
              result.Kanten[result.Anzahl].Start:= Knoten[i];
              result.Kanten[result.Anzahl].Ziel:= Knoten[Vorgaenger[i]];
            end;
     end;
end;

end.
