{ OOP-Deklaration der gngigen dynamischen Listenstrukturen        <U_adt.Pas> }
{                                                               vom 18.01.2002 }
{ Die Nachfolgend deklaraierten Listenstrukturen arbeiten objektorientiert     }
{ unter den folgenden Gesichtspunkten:                                         }
{                                                                              }
{   1. Die Inhaltsobjekte der Listenstrukturen sind vom allgmeinen Urobjekt    }
{      TElement. In der Oberflchen-Unit knnen beliebige Objekte von TElement }
{      abgeleitet werden. Denkbar sind z. B.:                                  }
{          TStringElement                                                      }
{          TIntegerElement                                                     }
{          TBoolElement                                                        }
{          ...                                                                 }
{                                                                              }
{   2. Die Listenstrukturen an sich sind als Klasse formuliert und somit       }
{      "erweiterbar". Beispiele finden sich im Grafikprojekt etc.              }
{                                                                              }
{   3. Die Datenstruktur TLinList ist "Zeigerfrei". Die Schler arbeiten also  }
{      nur noch mit der bekannten Syntax fr Objektorientierte Datenstrukturen.}
{      Ein new(TZeiger) wird damit zu TZeiger.Create.                          }
{                                                                              }
{ ----------------------------------------  Version 4.0 (Garmann) 18.01.2002 - }

unit U_adt;

interface

type
    TElement = class                        { Urobjekt fr Listenelemente      }
       private                              { jetzt Inhaltslos                 }
       Next: TElement;                      { Nachfolger                       }
       public
       constructor create;                  { Konstruktor wird verndert       }
                                            { Destruktor wird geerbt           }
     end;

     TLinList = class                           { Liste ber Urtyp TElement    }
       private
       sAnfang,                                 { "Zeiger" auf erstes/         }
       sAktuell: TElement;                      { aktuelles Element            }
       public
       constructor create;                      { setzt Kopf/Fuss auf NIL      }
       destructor destroy; override;            { lscht alle vorh. Elemente   }
       { Navigationsroutinen }
       procedure vor;                           { Aktuell einen weiter bewegen }
       procedure zurueck;                       { Aktuell einen zurck bewegen }
       procedure anAnfang;                      { Aktuell an den Anfang        }
       procedure ansEnde;                       { Aktuell an das Ende          }
       { Datensatzoperationen }
       procedure EinfuegenVor(e: TElement);     { Datensatz vor Aktuell einf.  }
       procedure EinfuegenNach(e: TElement);    { Datensatz nach Aktuell einf. }
       procedure Loeschen;                      { Aktuellen Datensatz loeschen }
       { Informationsroutinen }
       function Anzahl: integer;                { Anzahl der Datenstze        }
       function amEnde: boolean;                { Aktuell am Ende?             }
       function Leer: boolean;                  { Nzahl = 0?                   }
       function Aktuell: TElement;              { Aktuelles Element ausgeben   }
     end;


     TQueue = class                             { Queue ber Urtyp TElement    }
       private
       Kopf,                                    { "Zeiger" auf erstes/         }
       Fuss: TElement;                          { letztes Element              }
       public
       constructor create;                      { setzt Kopf/Fuss auf NIL      }
       destructor destroy; override;            { lscht alle vorh. Elemente   }
       procedure enqueue(i: TElement);          { neues Element in Schlange    }
       procedure dequeue;                       { vorderstes Element lschen   }
       function front: TElement;                { Rckgabe: vordersten Element }
       function empty: boolean;                 { Queue leer?                  }
     end;

     TStack = class                             { Stack ber Urtyp TElement    }
       private
       sTop: TElement;                          { "Zeiger" auf erstes Element  }
       public
       constructor create;                      { setzt sTop auf NIL           }
       destructor destroy; override;            { lscht alle vorh. Elemente   }
       procedure push(i: TElement);             { neues Element auf Stapel     }
       procedure pop;                           { oberstes Element lschen     }
       function top: TElement;                  { Rckgabe: oberstes Element   }
       function empty: boolean;                 { Stack leer?                  }
     end;

implementation

{ **** Routinen fr TElement **** }

constructor TElement.create;                { Konstruktor                      }
begin
  inherited create;                         { ererbten Konstruktor aufrufen    }
  Next:= NIL;                               { und Nachfolger initialisieren    }
end;


{ **** Routinen fr TLinList **** }

constructor TLinList.create;                { Konstruktor                      }
begin
  inherited create;                         { ererbten Konstruktor aufrufen    }
  sAnfang:= NIL;                            { und Zeiger auf erstes Element    }
                                            { mit NIL initialisieren           }
  sAktuell:= NIL;                           { und Zeiger auf aktuelles Element }
                                            { mit NIL initialisieren           }
end;

destructor TLinList.destroy;                { Destruktor                       }
begin
  while not Leer do Loeschen;               { alle Listenelemente lschen      }
  inherited destroy;                        { und schlielich den geerbten     }
                                            { Destruktor aufrufen              }
end;

procedure TLinList.vor;                     { einen Datensatz vor              }
begin
  if (not Leer) and (sAktuell.Next<>NIL)
    then sAktuell:= sAktuell.Next;
end;

procedure TLinList.zurueck;                 { einen Datensatz zurck           }
var h: TElement;
begin
  if (not Leer) and (sAktuell<>sAnfang)
    then begin
           h:= sAnfang;
           while h.Next<>sAktuell do h:= h.Next;
           sAktuell:= h;
         end;
end;

procedure TLinList.anAnfang;                { ersten Datensatz anwhlen        }
begin
  sAktuell:= sAnfang;
end;

procedure TLinList.ansEnde;                 { letzten Datensatz anwhlen       }
begin
  if not Leer
  then begin
         while sAktuell.Next<>NIL do vor;
       end;
end;

procedure TLinList.EinfuegenVor(e: TElement); { Neues element in Liste vor     }
                                              { Aktuell einfuegen              }
begin
  if Leer                                     { Wenn Liste noch leer           }
    then begin
           sAnfang:= e;                       { dann El. als erstes einreihen  }
           sAktuell:= e;
         end
    else if sAktuell=sAnfang                  { Aktuell steht auf erstem Elem. }
         then begin
                e.Next:= sAnfang;
                sAnfang:= e;
                sAktuell:= e;
              end
         else begin                           { Aktuell steht irgendwo sonst   }
                Zurueck;                      { erst ein Element zurck und    }
                EinfuegenNach(e);             { dann hinter Aktuell einfgen   }
              end;
end;

procedure TLinList.EinfuegenNach(e: TElement);{ Neues element in Liste vor     }
                                              { Aktuell einfuegen              }
begin
  if Leer                                     { Wenn Liste noch leer           }
    then begin
           sAnfang:= e;                       { dann El. als erstes einreihen  }
           sAktuell:= e;
         end
    else begin                                { sonst hinter Aktuell einreihen }
           e.Next:= sAktuell.Next;
           sAktuell.Next:= e;
           sAktuell:= e;
         end
end;

procedure TLinList.Loeschen;                  { Element aus Liste lschen      }
var h: TElement;
begin
  if not Leer                                 { nur wenn Liste nicht leer      }
  then begin
         if sAktuell=sAnfang                  { Wenn erstes Elem.gelscht wird,}
         then begin
                sAnfang:= sAnfang.Next;       { dann Anfang verschieben        }
                sAktuell.free;                { Aktuelles Element lschen      }
                sAktuell:= sAnfang;           { neues Aktuelles Element        }
              end
         else begin
                h:= sAktuell;                 { aktuelles Element merken       }
                zurueck;                      { zurueck gehen, um auszuketten  }
                sAktuell.next:= h.next;       { bisheriges akt. Elem. ausketten}
                h.free;                       { bisheriges akt. Elem. lschen  }
                if sAktuell.Next<>NIL         { Wenn nach Aktuell noch was ist,}
                then vor;                     { dann das nachfolgende nehemen  }
              end;
       end
end;

function TLinList.Aktuell: TElement;          { Rckgabe: aktuelles Element    }
begin
  Aktuell:= sAktuell                          { Rckgabe von Aktuell           }
end;

function TLinList.Leer: boolean;              { Liste leer?                    }
begin
  Leer:= (sAnfang=NIL);                       { Anfang = NIL => Liste leer     }
end;

function TLinList.Anzahl: integer;            { Elementanzahl der Liste        }
var anz: integer;
    h: TElement;
begin
  anz:= 0;                                    { Initialisierung                }
  if not Leer                                 { wenn Liste nicht leer ist,     }
  then begin
         h:= sAnfang;                         { vom Anfang bis                 }
         while h<>NIL                         { zum Ende (h=NIL)               }
         do begin
              h:= h.Next;                     { mit h die Listenelemente durch-}
              inc(anz);                       { laufen und Anzahl erhhen.     }
            end;
       end;
  Anzahl:= anz;                               { Funktionsergebnis-Rckgabe     }
end;

function TLinList.amEnde: boolean;
begin
  amEnde:= Leer or (sAktuell.Next=NIL);
end;



{ **** Routinen fr TQueue **** }

constructor TQueue.create;                  { Konstruktor                      }
begin
  inherited create;                         { ererbten Konstruktor aufrufen    }
  Kopf:= NIL;                               { und Zeiger auf erstes Element    }
                                            { mit NIL initialisieren           }
  Fuss:= NIL;                               { und Zeiger auf erstes Element    }
                                            { mit NIL initialisieren           }
end;

destructor TQueue.destroy;                  { Destruktor                       }
begin
  while not empty do dequeue;               { alle Queueelemente lschen       }
  inherited destroy;                        { und schlielich den geerbten     }
                                            { Destruktor aufrufen              }
end;

procedure TQueue.enqueue(i: TElement);      { Neues element in Schlange        }
begin
  if empty                                  { Wenn Schlange noch leer          }
    then begin
           Fuss:= i;                        { dann El. als erstes einreihen    }
           Kopf:= i;
         end
    else begin
           Fuss.Next:= i;                   { sonst Element hinten eingliedern }
           Fuss:= i;                        { Fuss-Zeiger weiterrcken         }
         end;
end;

procedure TQueue.dequeue;                   { Element aus Schlange lschen     }
var h: TElement;
begin
  if not empty                              { nur wenn Schlange nicht leer     }
  then begin
         h:= Kopf;                          { bisheriges erstes Element merken }
         Kopf:= Kopf.Next;                  { Nachfolger wird erstes Element   }
         h.destroy;                         { bisheriges erstes Elem. lschen  }
         if Kopf=NIL                        { war das gelschte El. das letzte }
           then Fuss:= NIL;                 {  dann auch Fuss auf NIL          }
       end
end;

function TQueue.front: TElement;            { Rckgabe: oberstes Element       }
begin
  front:= Kopf;                             
end;

function TQueue.empty: boolean;             { Queue leer?                      }
begin
  empty:= (Kopf=NIL);                       { Kopf = NIL => Queue leer         }
end;

{ **** Routinen fr TStack **** }

constructor TStack.create;                  { Konstruktor                      }
begin
  inherited create;                         { ererbten Konstruktor aufrufen    }
  sTop:= NIL;                               { und Zeiger auf erstes Element    }
                                            { mit NIL initialisieren           }
end;

destructor TStack.destroy;                  { Destruktor                       }
begin
  while not empty do pop;                   { alle Stackelemente lschen       }
  inherited destroy;                        { und schlielich den geerbten     }
                                            { Destruktor aufrufen              }
end;

procedure TStack.push(i: TElement);         { Neues element auf Stack          }
                                            { Das Element i ist bereits        }
                                            { erzeugt und initialisiert        }
                                            { was fehlt ist der Nachfolger     }
begin
  i.next:= sTop;                            { Neues Element wird vor das bis-  }
                                            { herige oberste Element gesetzt.  }
  sTop:= i;                                 { Also sTop -> neues Element       }
                                            { ACHTUNG: keine Kopie, sondern    }
                                            { der Zeiger sTop wird umgebogen.  }
end;

procedure TStack.pop;                       { Element vom Stack lschen        }
var h: TElement;
begin
  if not empty                              { nur wenn Stack nicht leer        }
  then begin
         h:= sTop;                          { bisheriges erstes Element merken }
         sTop:= sTop.next;                  { Nachfolger wird oberstes Element }
         h.destroy;                         { bisheriges erstes Elem. lschen  }
       end;
end;

function TStack.top: TElement;              { Rckgabe: oberstes Element       }
begin
  top:= sTop;
end;

function TStack.empty: boolean;             { Stack leer?                      }
begin
  empty:= (sTop=NIL);                       { sTop = NIL => Stack leer         }
end;


end.
